package id

import (
	"fmt"
	"reflect"
)

var registry = NewRegistry()

type ObjectHandler func(interface{}) *ObjectId

type Registry struct {
	handlers    map[reflect.Type]ObjectHandler
	descriptors map[string]Descriptor
}

func NewRegistry() *Registry {
	return &Registry{
		handlers:    make(map[reflect.Type]ObjectHandler),
		descriptors: make(map[string]Descriptor),
	}
}

// Register method adds a handler based on its type.
func (r *Registry) RegisterObjectHandler(t reflect.Type, h ObjectHandler) {
	r.handlers[t] = h
}

func (r *Registry) RegisterDescriptor(d Descriptor) {
	r.descriptors[d.Type()] = d
}

func (r *Registry) FromParts(s string) (*ObjectId, error) {
	parts := Split(s)

	for _, d := range r.descriptors {
		id := d.New()
		if err := id.FromParts(parts); err == nil {
			return &ObjectId{Descriptor: id}, nil
		}
	}

	return nil, fmt.Errorf("%w: %s", ErrInvalidID, s)
}

func (r *Registry) FromMap(m map[string]interface{}) (*ObjectId, error) {
	t, ok := m["type"].(string)
	if !ok {
		return nil, ErrTypeNotFound
	}

	if d, ok := r.descriptors[t]; ok {
		id := d.New()
		if err := id.FromMap(m); err != nil {
			return nil, err
		}
		return &ObjectId{Descriptor: id}, nil
	}

	return nil, fmt.Errorf("%s: %s", t, ErrInvalidID)
}

func (r *Registry) FromObject(v interface{}) (*ObjectId, error) {
	value := reflect.ValueOf(v)
	if v == nil || (value.Kind() == reflect.Ptr && value.IsNil()) {
		return nil, fmt.Errorf("object value is nil")
	}

	t := value.Type()
	if handler, ok := r.handlers[t]; ok {
		i := handler(v)
		if i == nil {
			panic(fmt.Sprintf("handler for %s returned nil", t))
		}
		return i, nil
	}
	return nil, fmt.Errorf("%s : %w", t, ErrHandlerNotFound)
}

func FromMap(m map[string]interface{}) (*ObjectId, error) {
	return registry.FromMap(m)
}

func FromString(s string) (*ObjectId, error) {
	return registry.FromParts(s)
}

type ObjectIdentifier interface {
	ObjectId() *ObjectId
}

func FromObject(v interface{}) (*ObjectId, error) {
	if id, ok := v.(ObjectIdentifier); ok {
		return id.ObjectId(), nil
	}
	return registry.FromObject(v)
}

func RegisterSystemIds(r *Registry) {
	r.RegisterDescriptor(&SpaceId{})
	r.RegisterDescriptor(&EnvironmentId{})
	r.RegisterDescriptor(&CollectionId{})
	r.RegisterDescriptor(&SchemaId{})
	r.RegisterDescriptor(&ItemId{})
	r.RegisterDescriptor(&RevisionId{})
	r.RegisterDescriptor(&FieldId{})
	r.RegisterDescriptor(&ClientId{})
	r.RegisterDescriptor(&RoleId{})
	r.RegisterDescriptor(&UserId{})
	r.RegisterDescriptor(&SystemId{})
	r.RegisterDescriptor(&ServiceId{})
	r.RegisterDescriptor(&OrganizationId{})
	r.RegisterDescriptor(&LocaleID{})
}

func GetRegistry() *Registry {
	return registry
}

func init() {
	RegisterSystemIds(registry)
}
