package setup

import (
	"context"
	"io/fs"

	"git.perx.ru/perxis/perxis-go/pkg/clients"
	"git.perx.ru/perxis/perxis-go/pkg/collections"
	"git.perx.ru/perxis/perxis-go/pkg/content"
	"git.perx.ru/perxis/perxis-go/pkg/errors"
	"git.perx.ru/perxis/perxis-go/pkg/items"
	"git.perx.ru/perxis/perxis-go/pkg/roles"
	"git.perx.ru/perxis/perxis-go/pkg/spaces"
	"go.uber.org/zap"
)

var (
	ErrInvalidSetupConfig = errors.New("invalid setup config")
)

// Setup реализует процесс настройки пространства. Указав необходимые требования к конфигурации пространства можно
// выполнить процесс установки, проверки и удаления требований.
type Setup struct {
	SpaceID       string
	EnvironmentID string

	content *content.Content

	force  bool
	remove bool

	waitSpaceAvailable bool
	//attempts          uint
	//delay        time.Duration

	errors []error
	logger *zap.Logger

	*Config
}

func NewSetup(content *content.Content, spaceID, environmentID string, logger *zap.Logger) *Setup {
	if logger == nil {
		logger = zap.NewNop()
	}

	logger = logger.With(zap.String("Space", spaceID), zap.String("Environment", environmentID))

	return &Setup{
		SpaceID:            spaceID,
		EnvironmentID:      environmentID,
		content:            content,
		logger:             logger,
		waitSpaceAvailable: true,
		Config:             NewConfig(),
	}
}

func (s *Setup) WithConfig(cfg *Config) *Setup {
	setup := *s
	setup.Config = cfg
	return &setup
}

func (s *Setup) WithForce(force bool) *Setup {
	setup := *s
	setup.force = force
	return &setup
}

func (s *Setup) IsForce() bool {
	return s.force
}

func (s *Setup) WithRemove(remove bool) *Setup {
	setup := *s
	setup.remove = remove
	return &setup
}

func (s *Setup) IsRemove() bool {
	return s.remove
}

func (s *Setup) WithoutWaitSpace() *Setup {
	setup := *s
	setup.waitSpaceAvailable = false
	return &setup
}

func (s *Setup) HasErrors() bool {
	return len(s.errors) > 0
}

func (s *Setup) AddError(err error) {
	s.errors = append(s.errors, err)
}

func (s *Setup) Errors() []error {
	return s.errors
}

func (s *Setup) Error() error {
	return errors.WithErrors(ErrInvalidSetupConfig, s.errors...)
}

// Install выполняет установку необходимых требований
func (s *Setup) Install(ctx context.Context) error {
	if s.waitSpaceAvailable {
		if err := spaces.WaitSpaceAvailable(ctx, s.content.Spaces, s.SpaceID); err != nil {
			return err
		}
	}
	if err := s.InstallRoles(ctx); err != nil {
		return err
	}
	if err := s.InstallClients(ctx); err != nil {
		return err
	}
	if err := s.InstallCollections(ctx); err != nil {
		return err
	}
	if err := s.InstallItems(ctx); err != nil {
		return err
	}
	return nil
}

// Check выполняет проверку требований
func (s *Setup) Check(ctx context.Context) error {
	if s.waitSpaceAvailable {
		if err := spaces.WaitSpaceReadable(ctx, s.content.Spaces, s.SpaceID); err != nil {
			return err
		}
	}
	if err := s.CheckRoles(ctx); err != nil {
		return err
	}
	if err := s.CheckClients(ctx); err != nil {
		return err
	}
	if err := s.CheckCollections(ctx); err != nil {
		return err
	}
	if err := s.CheckItems(ctx); err != nil {
		return err
	}
	return nil
}

// Uninstall выполняет удаление установленных раннее требований
func (s *Setup) Uninstall(ctx context.Context) error {
	if s.waitSpaceAvailable {
		if err := spaces.WaitSpaceAvailable(ctx, s.content.Spaces, s.SpaceID); err != nil {
			return err
		}
	}
	// В случае если необходимо удалить данные удаляем все что создано при установке расширения
	if err := s.UninstallClients(ctx); err != nil {
		return err
	}
	if err := s.UninstallRoles(ctx); err != nil {
		return err
	}
	if err := s.UninstallCollections(ctx); err != nil {
		return err
	}
	if err := s.UninstallItems(ctx); err != nil {
		return err
	}
	return nil
}

// Deprecated: use Config
func (s *Setup) Load(fsys fs.FS) *Setup {
	var err error
	s.Config, err = s.Config.Load(fsys)
	if err != nil {
		s.AddError(err)
	}
	return s
}

// Deprecated: use Config instead
// AddCollections добавляет требования к настройке коллекций в пространстве
func (s *Setup) AddCollections(collections []*collections.Collection, opt ...CollectionsOption) *Setup {
	for _, col := range collections {
		s.AddCollection(col, opt...)
	}
	return s
}

// Deprecated: use Config instead
// AddCollection добавляет требование к настройке коллекции в пространстве
func (s *Setup) AddCollection(collection *collections.Collection, opt ...CollectionsOption) *Setup {
	if _, err := s.Config.AddCollection(collection, opt...); err != nil {
		s.AddError(err)
	}
	return s
}

// Deprecated: use Config instead
// AddRoles добавляет требования к настройке ролей в пространстве
func (s *Setup) AddRoles(roles []*roles.Role, opt ...RolesOption) *Setup {
	s.Config = s.Config.AddRoles(roles, opt...)
	return s
}

// Deprecated: use Config instead
// AddRole добавляет требования к настройке элементов в пространстве
func (s *Setup) AddRole(role *roles.Role, opt ...RolesOption) *Setup {
	s.Config = s.Config.AddRole(role, opt...)
	return s
}

// Deprecated: use Config instead
// AddItem добавляет требования к настройке элементов в пространстве
func (s *Setup) AddItem(item *items.Item, opt ...ItemsOption) *Setup {
	s.Config = s.Config.AddItem(item, opt...)
	return s
}

// Deprecated: use Config instead
// AddItems добавляет требования к настройке элементов в пространстве
func (s *Setup) AddItems(items []*items.Item, opt ...ItemsOption) *Setup {
	s.Config = s.Config.AddItems(items, opt...)
	return s
}

// Deprecated: use Config instead
// AddClient добавляет требования к настройке элементов в пространстве
func (s *Setup) AddClient(client *clients.Client, opt ...ClientsOption) *Setup {
	s.Config = s.Config.AddClient(client, opt...)
	return s
}

// Deprecated: use Config instead
// AddClients добавляет требования к настройке элементов в пространстве
func (s *Setup) AddClients(clients []*clients.Client, opt ...ClientsOption) *Setup {
	s.Config = s.Config.AddClients(clients, opt...)
	return s
}
